/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication;

import org.jflux.api.core.util.ListMergeAdapter;
import java.util.List;
import org.jflux.api.core.Source;
import org.jflux.api.core.chain.SourceChain;
import org.jflux.api.core.node.DefaultProcessorNode;
import org.jflux.api.core.util.SourceAdapter;
import org.jflux.api.data.blend.DataCollector;
import org.jflux.api.data.blend.DataCollector.DefaultCollector;
import org.jflux.avrogen.EventRecord;
import org.osgi.framework.BundleContext;
import org.robokind.api.common.osgi.ServiceClassListener;
import org.robokind.demo.robot.replication.lifecycle.EventBufferUtils.EventBuffer;

/**
 *
 * @author Matthew Stevenson
 */
public class EventListCollector extends 
        DefaultProcessorNode<Object,List<EventRecord>> implements 
        DataCollector<EventRecord,List<EventRecord>> {
    private DataCollector<List<EventRecord>,List<List<EventRecord>>> myEventListCollector;
    private DefaultCollector<EventRecord> myEventCollector;
    private Source<List<EventRecord>> myEventsSource;
    
    public static EventListCollector buildDefault() {
        DefaultCollector<List<EventRecord>> col = 
                new DefaultCollector<List<EventRecord>>();
        EventListCollector c = new EventListCollector(
                col, SourceChain.builder(col)
                    .attach(new ListMergeAdapter<EventRecord>()).done());
        c.start();
        return c;
    }
    
    EventListCollector(
            DataCollector<List<EventRecord>,List<List<EventRecord>>> listCollector, 
            Source<List<EventRecord>> adaptedCollector) {
        super(new SourceAdapter<Object, List<EventRecord>>(adaptedCollector));
        myEventListCollector = listCollector;
        myEventsSource = adaptedCollector;
        myEventCollector = new DefaultCollector<EventRecord>();
        myEventListCollector.addSource(myEventCollector);
    }
    
    public Source<List<EventRecord>> getRecordSource(){
        return myEventsSource;
    }
    
    public void addListSource(Source<List<EventRecord>> source) {
        myEventListCollector.addSource(source);
    }
    
    @Override
    public void addSource(Source<EventRecord> source) {
        myEventCollector.addSource(source);
    }

    public void removeListSource(Source<List<EventRecord>> source) {
        myEventListCollector.addSource(source);
    }
    
    @Override
    public void removeSource(Source<EventRecord> source) {
        myEventCollector.addSource(source);
    }

    @Override
    public List<EventRecord> getValue() {
        return myEventsSource.getValue();
    }
    
    public static class EventBufferTracker extends ServiceClassListener<EventBuffer> {
        private EventListCollector myCollector;
        
        public EventBufferTracker(BundleContext context, 
                String filterString, EventListCollector collector){
            super(EventBuffer.class, context, filterString);
            if(collector == null){
                throw new NullPointerException();
            }
            myCollector = collector;
        }

        @Override
        protected void addService(EventBuffer t) {
            if(t == null || t.getRecordValues() == null){
                return;
            }
            myCollector.addListSource(t.getRecordValues());
        }

        @Override
        protected void removeService(EventBuffer t) {
            if(t == null || t.getRecordValues() == null){
                return;
            }
            myCollector.removeListSource(t.getRecordValues());
        }
    }
}
