/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication.lifecycle;
import org.apache.avro.specific.SpecificRecordBase;
import org.jflux.api.core.event.Event;
import org.jflux.api.core.event.Header;
import org.robokind.demo.robot.replication.RoutingDecoderNode.EventRouteLink;
import java.io.ByteArrayInputStream;
import org.robokind.demo.robot.replication.RoutingDecoderNode.DecoderLink;
import org.jflux.api.core.node.ProducerNode;
import org.robokind.api.common.lifecycle.config.GenericLifecycle.DependencyChange;
import org.jflux.impl.transport.jms.MessageUnpacker;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import javax.jms.BytesMessage;
import javax.jms.Destination;
import javax.jms.Session;
import org.jflux.api.core.Adapter;
import org.jflux.api.core.Listener;
import org.jflux.api.core.chain.AdapterChain;
import org.jflux.api.core.config.Configuration;
import org.jflux.api.core.node.ConsumerNode;
import org.jflux.api.core.node.ProcessorNode;
import org.jflux.api.core.node.chain.NodeChainBuilder;
import org.jflux.api.core.util.IteratorNode;
import org.jflux.api.encode.BytesUtils;
import org.jflux.avrogen.EventListRecord;
import org.jflux.avrogen.EventRecord;
import org.jflux.impl.encode.avro.AvroDecoder;
import org.jflux.impl.transport.jms.JMSMessageReceiver;
import org.robokind.demo.robot.replication.EventDecoder.EventListDecoder;
import org.robokind.demo.robot.replication.RoutingDecoderNode;

import org.robokind.demo.robot.replication.chain.NodeLinkLifecycleUtils;
import static org.jflux.impl.encode.avro.SerializationConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKDependencyConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKLifecycleConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKManagedGroupConfigUtils.*;

/**
 *
 * @author Matthew Stevenson
 */
public class ReplicationReceiverLifecycleUtils {   
    private final static String theDestination = "destination";
    private final static String theSession = "session";
    private final static String theRouter = "router";
    private final static String theEncodeConfig = "encodeConfig";
    private final static String theConsumer = "consumer";
    
    public static List<Configuration<String>> getLifecycleConfigs(
            String routerNodeIdKey, String routerNodeId,
            String eventListDecoderIdKey, String eventListDecoderId,
            String receiverIdKey, String receiverId,
            String destinationIdKey, String destinationId, 
            String sessionIdKey, String sessionId,
            String receiverListDecoderLinkId,
            String listDecoderRouterLinkId){
        return Arrays.asList(
                buildEventListDecoderNodeLifecycleConfig(
                eventListDecoderIdKey, eventListDecoderId),
                buildRoutingDecoderLifecycleConfig(routerNodeIdKey, routerNodeId),
                buildReceiverLifecycleConfig(receiverIdKey, receiverId, 
                        destinationIdKey, destinationId, 
                        sessionIdKey, sessionId),
                buildReceiverListDecoderLinkLifecycleConfigs(
                        receiverListDecoderLinkId, 
                        receiverIdKey, receiverId, 
                        eventListDecoderIdKey, eventListDecoderId),
                buildListDecoderRouterLinkLifecycleConfigs(
                        listDecoderRouterLinkId, 
                        eventListDecoderIdKey, eventListDecoderId, 
                        routerNodeIdKey, routerNodeId));
    }
    
    public static Configuration<String> buildReceiverListDecoderLinkLifecycleConfigs(
            String receiverListDecoderLinkId,
            String receiverIdKey, String receiverId,
            String listDecoderIdKey, String listDecoderId){
        return NodeLinkLifecycleUtils.producerProcessorLinkLifecycleConfig(
                        receiverListDecoderLinkId, null, 
                        receiverIdKey, receiverId, null, 
                        listDecoderIdKey, listDecoderId, null);
    }
    
    public static Configuration<String> buildListDecoderRouterLinkLifecycleConfigs(
            String listDecoderRouterLinkId,
            String listDecoderIdKey, String listDecoderId,
            String routerIdKey, String routerId){
        return NodeLinkLifecycleUtils.processorConsumerLinkLifecycleConfig(
                        listDecoderRouterLinkId, null,
                        listDecoderIdKey, listDecoderId, null, 
                        routerIdKey, routerId, null);
    }
    
    public static Configuration<String> buildEventListDecoderNodeLifecycleConfig(
            String eventListDecoderIdKey, String eventListDecoderId){
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{ProcessorNode.class.getName()}, 
                eventListDecoderIdKey, eventListDecoderId, 
                null, Collections.EMPTY_LIST, 
                new Adapter<Map<String,Object>,ProcessorNode>() {
                    @Override
                    public ProcessorNode<BytesMessage,EventRecord> adapt(Map<String,Object> a) {
                        Adapter<BytesMessage, List<EventRecord>> adapter =
                                AdapterChain.builder(new MessageUnpacker())
                                    .attach(AvroDecoder.buildSpecificBinaryDecoder(
                                            EventListRecord.class))
                                    .attach(new EventListDecoder()).done();
                        ProcessorNode<BytesMessage,EventRecord> n =
                                NodeChainBuilder.build(adapter)
                                        .attach(new IteratorNode<EventRecord>())
                                        .getProcessorChain();
                        n.start();
                        return n;
                    }},
                new Listener<ProcessorNode>() {
                    @Override public void handleEvent(ProcessorNode input) {
                        input.stop();
                    }
                }));
    }
    
    public static Configuration<String> buildReceiverLifecycleConfig(
            String receiverIdKey, String receiverId,
            String destinationIdKey, String destinationId, 
            String sessionIdKey, String sessionId){
        List<Configuration<String>> deps = Arrays.asList(
            buildLifecycleDependencyConfig(theDestination, Destination.class, 
                destinationIdKey, destinationId, null, 
                new Listener<DependencyChange<JMSMessageReceiver,Destination>>(){
                    @Override public void handleEvent(
                            DependencyChange<JMSMessageReceiver,Destination> input) {
                        if(input.getService() != null){
                            input.getService().setDestination(input.getDependency());
                        }
                    }}),
            buildLifecycleDependencyConfig(theSession, Session.class, 
                sessionIdKey, sessionId, null, 
                new Listener<DependencyChange<JMSMessageReceiver,Session>>(){
                    @Override public void handleEvent(
                            DependencyChange<JMSMessageReceiver,Session> input) {
                        if(input.getService() != null){
                            input.getService().setSession(input.getDependency());
                        }
                    }}));
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{ProducerNode.class.getName(),JMSMessageReceiver.class.getName()}, 
                receiverIdKey, receiverId, null, deps, 
                new Adapter<Map<String, Object>, JMSMessageReceiver>() {
                    @Override public JMSMessageReceiver adapt(Map<String, Object> a) {
                        Destination d = (Destination)a.get(theDestination);
                        Session s = (Session)a.get(theSession);
                        JMSMessageReceiver r = new JMSMessageReceiver(s, d);
                        r.start();
                        return r;
                    }},
                new Listener<JMSMessageReceiver>() {
                    @Override public void handleEvent(JMSMessageReceiver input) {
                        input.stop();
                    }
                }));        
    }
    
    public static Configuration<String> buildRoutingDecoderLifecycleConfig(
            String nodeIdKey, String nodeId){
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{ConsumerNode.class.getName(), RoutingDecoderNode.class.getName()}, 
                nodeIdKey, nodeId, null, Collections.EMPTY_LIST, 
                new Adapter<Map<String, Object>, RoutingDecoderNode>() {
                    @Override public RoutingDecoderNode adapt(Map<String, Object> a) {
                        RoutingDecoderNode r = RoutingDecoderNode.makeDefault();
                        r.start();
                        return r;
                    }
                })); 
    }
    
    
    public static <T, R extends SpecificRecordBase> Configuration<String> buildDecoderRouteLinkLifecycleConfig(
            String linkIdKey, String linkId, 
            String routerIdKey, String routerId, 
            String encodeConfigIdKey, String encodeConfigId){
        List<Configuration<String>> deps = Arrays.asList(
            buildLifecycleDependencyConfig(theRouter, RoutingDecoderNode.class, 
                routerIdKey, routerId, null, 
                new Listener<DependencyChange<DecoderLink,RoutingDecoderNode>>(){
                    @Override public void handleEvent(
                            DependencyChange<DecoderLink,RoutingDecoderNode> input) {
                        if(input.getService() != null){
                            input.getService().setRoutingDecoder(input.getDependency());
                        }
                    }}),
            buildLifecycleDependencyConfig(theEncodeConfig, Configuration.class, 
                encodeConfigIdKey, encodeConfigId, null, 
                new Listener<DependencyChange<DecoderLink,Configuration>>(){
                    @Override public void handleEvent(
                            DependencyChange<DecoderLink,Configuration> input) {
                        if(input.getService() == null){
                            return;
                        }
                        Configuration<String> c = input.getDependency();
                        if(c == null){
                            input.getService().setDecoder(null, null);
                            return;
                        }
                        Class<T> clazz = c.getPropertyValue(CONF_MESSAGE_CLASS);
                        Adapter<ByteArrayInputStream,T> decoder = c.getPropertyValue(CONF_DECODING_ADAPTER);
                        input.getService().setDecoder(clazz.toString(), decoder);
                    }}));
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{DecoderLink.class.getName()}, 
                linkIdKey, linkId, null, deps, 
                new Adapter<Map<String, Object>, DecoderLink>() {
                    @Override public DecoderLink adapt(Map<String, Object> a) {
                        RoutingDecoderNode r = (RoutingDecoderNode)a.get(theRouter);
                        Configuration<String> c = (Configuration)a.get(theEncodeConfig);
                        Class<T> clazz = c.getPropertyValue(CONF_MESSAGE_CLASS);
                        Class<R> recClazz = c.getPropertyValue(CONF_OUTPUT_CLASS);
                        Adapter<R,T> decoder = c.getPropertyValue(CONF_DECODING_ADAPTER);
                        return new DecoderLink<T>(r, clazz.toString(), 
                                AdapterChain.builder(new BytesUtils.BufferToInputStream())
                                .attach(AvroDecoder.buildByteStreamDecoder(recClazz))
                                .attach(decoder).done());
                    }},
                new Listener<DecoderLink>() {
                    @Override public void handleEvent(DecoderLink input) {
                        input.unlink();
                    }
                }));
    }
    
    public static <T> Configuration<String> buildEventRouteLinkLifecycleConfig(
            String linkIdKey, String linkId,
            String routerIdKey, String routerId, 
            String consumerIdKey, String consumerId, 
            final String routeKey){
        List<Configuration<String>> deps = Arrays.asList(
            buildLifecycleDependencyConfig(theRouter, RoutingDecoderNode.class, 
                routerIdKey, routerId, null, 
                new Listener<DependencyChange<EventRouteLink,RoutingDecoderNode>>(){
                    @Override public void handleEvent(
                            DependencyChange<EventRouteLink,RoutingDecoderNode> input) {
                        if(input.getService() != null){
                            input.getService().setRoutingDecoder(input.getDependency());
                        }
                    }}),
            buildLifecycleDependencyConfig(theConsumer, ConsumerNode.class, 
                consumerIdKey, consumerId, null, 
                new Listener<DependencyChange<EventRouteLink,ConsumerNode>>(){
                    @Override public void handleEvent(
                            DependencyChange<EventRouteLink,ConsumerNode> input) {
                        if(input.getService() != null){
                            input.getService().setRoute(routeKey, 
                                    input.getDependency() == null ? null : 
                                    input.getDependency().getListener());
                        }
                    }}));
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{EventRouteLink.class.getName()}, 
                linkIdKey, linkId, null, deps, 
                new Adapter<Map<String, Object>, EventRouteLink>() {
                    @Override public EventRouteLink adapt(Map<String, Object> a) {
                        RoutingDecoderNode r = (RoutingDecoderNode)a.get(theRouter);
                        ConsumerNode<Event<Header,T>> c = 
                                (ConsumerNode<Event<Header,T>>)a.get(theConsumer);
                        return new EventRouteLink<T>(r, routeKey, c.getListener());
                    }},
                new Listener<EventRouteLink>() {
                    @Override public void handleEvent(EventRouteLink input) {
                        input.unlink();
                    }
                }));
    }
}
