/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication.lifecycle;

import org.robokind.api.common.lifecycle.config.GenericLifecycle.DependencyChange;
import javax.jms.BytesMessage;
import org.jflux.api.core.node.ConsumerNode;
import org.jflux.api.core.playable.BasicPlayable;
import java.io.ByteArrayOutputStream;
import org.jflux.api.core.chain.AdapterChain;
import org.jflux.api.encode.EncodeRequest;
import org.jflux.avrogen.EventListRecord;
import org.jflux.avrogen.EventRecord;
import org.jflux.impl.encode.avro.AvroEncoder;
import org.jflux.impl.messaging.JMSAvroUtils.ByteOutputStreamFactory;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import javax.jms.Destination;
import javax.jms.Session;
import org.jflux.api.core.Adapter;
import org.jflux.api.core.Listener;
import org.jflux.api.core.config.Configuration;
import org.jflux.api.core.node.DefaultProcessorNode;
import org.jflux.api.core.node.ProcessorNode;
import org.jflux.api.core.node.chain.NodeChainBuilder;
import org.jflux.api.core.util.MapAdapter.MapValueAdapter;
import org.jflux.api.core.util.Repeater;
import org.jflux.avrogen.HeaderRecord;
import org.jflux.impl.encode.avro.SerializationConfigUtils;
import org.jflux.impl.transport.jms.JMSMessageSender;
import org.jflux.impl.transport.jms.MessageHeaderAdapter;
import org.jflux.impl.transport.jms.MessagePacker;
import org.robokind.demo.robot.replication.EventEncoder;
import static org.robokind.api.common.lifecycle.config.RKDependencyConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKLifecycleConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKManagedGroupConfigUtils.*;

/**
 *
 * @author Matthew Stevenson
 */
public class ReplicationSenderLifecycleUtils {
    private final static String theEncodeConfig = "encodeConfig";    
    private final static String theDestination = "destination";
    private final static String theSession = "session";
    
    private final static String theMsgPacker = "msgPacker";
    private final static String theMsgHeader = "msgHeader";
    private final static String theMsgSender = "msgSender";
    
    private final static String theMsgPackerIdKey = "messagePackerId";
    private final static String theMsgHeaderIdKey = "messageHeaderAdapterId";
    private final static String theMsgSenderIdKey = ReplicationSenderGroupLifecycleUtils.REPLICATION_SENDER_ID;
    
    public final static String PROP_REPLICATION_SENDER_NODE_ID = "replicationNodeSenderId";
    
    public final static String DEFAULT_CONTENT_TYPE = "application/avro";
    
    public static List<Configuration<String>> getLifecycleConfigs(
            String encodeConfigIdKey, String encodeConfigId, 
            String destinationIdKey, String destinationId, 
            String sessionIdKey, String sessionId, 
            String msgPackerId, String msgHeaderId, 
            String replicationNodeId, String senderId){
        return Arrays.asList(
            buildMsgPackerLifecycleConfig(sessionIdKey, sessionId, msgPackerId),
            buildMsgHeaderLifecycleConfig(encodeConfigIdKey, encodeConfigId, msgHeaderId),
            buildSenderLifecycleConfig(destinationIdKey, destinationId, 
                sessionIdKey, sessionId, senderId),
            buildReplicationSenderLifecycleConfig(
                replicationNodeId, msgPackerId, msgHeaderId, senderId));
    }
    
    public static Adapter<List<EventRecord>,EventListRecord> eventListAdapter(
            String eventType, String sourceRef, Map<String,String> props){
        return new EventEncoder.EventListEncoder(
                new EventEncoder.HeaderRecordSource(eventType, sourceRef, props));
    }
    
    public static Adapter<EventListRecord,ByteArrayOutputStream> listByteAdapter(){
        return new AdapterChain(EncodeRequest.factory(new ByteOutputStreamFactory()), 
                new AvroEncoder<EventListRecord, ByteArrayOutputStream>(
                        EventListRecord.class, EventListRecord.SCHEMA$, false));
    }
    
    public static Configuration<String> buildMsgPackerLifecycleConfig(
            String sessionIdKey, String sessionId, String msgPackerId){
        Configuration<String> depConf = buildLifecycleDependencyConfig(
                theSession, Session.class, sessionIdKey, sessionId, null, null);
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{Adapter.class.getName(), MessagePacker.class.getName()}, 
                theMsgPackerIdKey, msgPackerId, null, Arrays.asList(depConf), 
                new MapValueAdapter(theSession, new Adapter<Session,MessagePacker>() {
                    @Override
                    public MessagePacker adapt(Session a) {
                        return new MessagePacker(a);
                    }})));
    }
    
    public static Configuration<String> buildMsgHeaderLifecycleConfig(
            String encodeConfigIdKey, String encodeConfigId, String msgHeaderId){
        Configuration<String> depConf = buildLifecycleDependencyConfig(
                theEncodeConfig, Configuration.class, 
                encodeConfigIdKey, encodeConfigId, null, null);
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{Adapter.class.getName(), MessageHeaderAdapter.class.getName()}, 
                theMsgHeaderIdKey, msgHeaderId, null, Arrays.asList(depConf), 
                new MapValueAdapter(theEncodeConfig, new Adapter<Configuration,MessageHeaderAdapter>() {
                    @Override
                    public MessageHeaderAdapter adapt(Configuration a) {
                        String ct = (String)a.getPropertyValue(
                                SerializationConfigUtils.CONF_CONTENT_TYPE);
                        ct = ct == null ? DEFAULT_CONTENT_TYPE : ct;
                        return new MessageHeaderAdapter(ct);
                    }})));
    }
    
    public static Configuration<String> buildSenderLifecycleConfig(
            String destinationIdKey, String destinationId, 
            String sessionIdKey, String sessionId, String senderId){
        List<Configuration<String>> deps = Arrays.asList(
            buildLifecycleDependencyConfig(theDestination, Destination.class, 
                destinationIdKey, destinationId, null, 
                new Listener<DependencyChange<JMSMessageSender,Destination>>(){
                    @Override public void handleEvent(
                            DependencyChange<JMSMessageSender,Destination> input) {
                        if(input.getService() != null){
                            input.getService().setDestination(input.getDependency());
                        }
                    }}),
            buildLifecycleDependencyConfig(theSession, Session.class, 
                sessionIdKey, sessionId, null, 
                new Listener<DependencyChange<JMSMessageSender,Session>>(){
                    @Override public void handleEvent(
                            DependencyChange<JMSMessageSender,Session> input) {
                        if(input.getService() != null){
                            input.getService().setSession(input.getDependency());
                        }
                    }})
        );
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{Listener.class.getName(),JMSMessageSender.class.getName()}, 
                theMsgSenderIdKey, senderId, null, deps, 
                new Adapter<Map<String, Object>, JMSMessageSender>() {
                    @Override public JMSMessageSender adapt(Map<String, Object> a) {
                        Destination d = (Destination)a.get(theDestination);
                        Session s = (Session)a.get(theSession);
                        return new JMSMessageSender(s, d);
                    }
                }));        
    }
    
    public static Configuration<String> buildReplicationSenderLifecycleConfig(
            final String replicationNodeId, String msgPackerId, 
            String msgHeaderId, String senderId){
        List<Configuration<String>> deps = Arrays.asList(
            buildLifecycleDependencyConfig(theMsgPacker, MessagePacker.class, 
                theMsgPackerIdKey, msgPackerId, null, null),
            buildLifecycleDependencyConfig(theMsgHeader, MessageHeaderAdapter.class, 
                theMsgHeaderIdKey, msgHeaderId, null, null),
            buildLifecycleDependencyConfig(theMsgSender, JMSMessageSender.class, 
                theMsgSenderIdKey, senderId, null, null)
        );
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{ConsumerNode.class.getName(),ReplicationSender.class.getName()}, 
                PROP_REPLICATION_SENDER_NODE_ID, replicationNodeId, null, deps, 
                new Adapter<Map<String, Object>, ReplicationSender>() {
                    @Override public ReplicationSender adapt(Map<String, Object> a) {
                        MessagePacker msgPacker = (MessagePacker)a.get(theMsgPacker);
                        MessageHeaderAdapter headerAdapter = 
                                (MessageHeaderAdapter)a.get(theMsgHeader);
                        JMSMessageSender sender = (JMSMessageSender)a.get(theMsgSender);
                        ReplicationSender s = new ReplicationSender(
                                replicationNodeId, 
                                EventListRecord.class.getSimpleName(), 
                                null, msgPacker, headerAdapter, sender);
                        s.start();
                        return s;
                    }
                },
                new Listener<ReplicationSender>() {
                    @Override public void handleEvent(ReplicationSender input) {
                        input.stop();
                    }}));        
    }
    
    private static class ReplicationSender extends 
            BasicPlayable implements ConsumerNode<List<EventRecord>> {
        private Adapter<List<EventRecord>,EventListRecord> myEventListAdapter;
        private Adapter<EventListRecord,ByteArrayOutputStream> myEncoder;
        private Adapter<ByteArrayOutputStream,BytesMessage> myMessagePacker;
        private Adapter<BytesMessage,BytesMessage> myMessageHeaderAdapter;
        private Listener<BytesMessage> myMessageSender;

        private Listener<List<EventRecord>> mySenderChain;

        public ReplicationSender(String replicationNodeId, String eventListType,
                Map<String,String> listProps, MessagePacker msgPacker,
                MessageHeaderAdapter headerAdapter, JMSMessageSender sender){
            myEventListAdapter = eventListAdapter(eventListType, replicationNodeId, listProps);
            myEncoder = listByteAdapter();
            myMessagePacker = msgPacker;
            myMessageHeaderAdapter = headerAdapter;
            myMessageSender = sender;
            ProcessorNode<List<EventRecord>,BytesMessage> a = 
                    new DefaultProcessorNode<List<EventRecord>, BytesMessage>(
                            AdapterChain.builder(myEventListAdapter)
                                    .attach(myEncoder)
                                    .attach(myMessagePacker)
                                    .attach(myMessageHeaderAdapter).done());
            ConsumerNode n = NodeChainBuilder.build(a)
                    .getConsumerChain(myMessageSender);
            Repeater<List<EventRecord>> repeater = 
                    new Repeater<List<EventRecord>>();
            repeater.addListener(n.getListener());
            //repeater.addListener(new LoggingListener());
            mySenderChain = repeater;
            n.start();
        }

        @Override
        public Listener<List<EventRecord>> getListener() {
            return mySenderChain;
        }
    }
    
    static class LoggingListener implements Listener<List<EventRecord>> {
        
        @Override 
        public void handleEvent(List<EventRecord> input) {
            if(input == null || input.isEmpty()){
                System.out.println("No events to send");
                return;
            }
            StringBuilder sb = new StringBuilder("Sending Events:");
            for(EventRecord r : input){
                addEventString(r, sb);
            }
            System.out.println(sb.toString());
        }

        private void addEventString(EventRecord r, StringBuilder sb) {
            if (r == null) {
                sb.append("\nnull");
                return;
            }
            HeaderRecord h = r.getHeader();
            if (h == null || h.getEventType() == null) {
                sb.append("\nunknown");
                return;
            }
            sb.append("\n").append(h.getEventType());
            if(h.getTimestamp() != null){
                sb.append(", event time: ")
                        .append(h.getTimestamp().toString());
            }
        }
    }
}
